/************************************************************************/
/*                                                                      */
/*      Tile Pages WE - Generic WebExtension - Container Page           */
/*                                                                      */
/*      Javascript for Container Page                                   */
/*                                                                      */
/*      Last Edit - 10 Jan 2019                                         */
/*                                                                      */
/*      Copyright (C) 2016-2019 DW-dev                                  */
/*                                                                      */
/*      Distributed under the GNU General Public License version 2      */
/*      See LICENCE.txt file and http://www.gnu.org/licenses/           */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Refer to Google Chrome developer documentation:                     */
/*                                                                      */
/*  https://developer.chrome.com/extensions/messaging                   */
/*                                                                      */
/*  https://developer.chrome.com/extensions/runtime                     */
/*  https://developer.chrome.com/extensions/storage                     */
/*                                                                      */
/************************************************************************/

"use strict";

/************************************************************************/

/* Global variables */

var homePage = "";

var showNavBar,setTitle;
var homePage;
var syncActive;

var lastActiveElement = null;

var layoutTabId = -1;  /* updated from background.js */
var layoutTabIncognito = false;  /* updated from background.js */

var tileBorder = 2;
var splitterSize = 6;
var splitterSnap = 20;
var navbarHeight = 32;

var panelCount = 0;
var panelSeqNo = 0;
var splitterSeqNo = 0;

var rootTile = null;

var activeTile = null;
var viewTiled = true;
var syncScroll = false;

var downTile = null;
var downSplitter = 0;
var lastScreenX = 0;
var lastScreenY = 0;
var moveTimeout = null;

var urlbarFocused = false;

var urlbarSuggestions = new Array();

var dragText = "";

/************************************************************************/

/* Tile object constructor */

function Tile()
{
    this.parentTile = null;
    this.type = "panel";  /* panel, hsplit or vsplit */
    
    /* Tile position and size*/
    
    this.left = 0;
    this.top = 0;
    this.width = 0;
    this.height = 0;
    this.ratio = 1;
    this.expandRatio = 1;
    
    /* Panel tile */
    
    this.panelId = -1;  /* panel for tile, -1 = no panel */
    
    this.scrollX = 0;  /* last scroll X */
    this.scrollY = 0;  /* last scroll Y */
    this.scrollInit = false;  /* initialize scroll position on first page load after opening layout */
    
    this.frameId = -1;
    this.scriptLoaded = false;
    
    this.historyIndex = -1;
    this.historyUrl = new Array();
    this.historyTitle = new Array();
    
    /* Split tile */
    
    this.childTiles = new Array();
    this.splitterIds = new Array();
}

/************************************************************************/

/* Listener for container page load */

document.addEventListener("DOMContentLoaded",onLoadPage,false);

/************************************************************************/

/* Initialize on page load */

function onLoadPage()
{
    document.body.onbeforeunload =
    function confirmExit(event)
    {
        /* Can be reload, navigation, tab close or window close */
        
        return "confirm exit";  /* any string shows default message */
    };
    
    chrome.storage.local.get(null,
    function(object)
    {
        var i,datalist,option;
        
        showNavBar = object["options-shownavbar"];
        setTitle = object["options-settitle"];
        
        homePage = object["options-homepage"];
        
        syncActive = object["options-syncactive"];
        
        homePage = homePage.replace(/\/$/,"");
        
        /* Initialize url bar suggestions*/
        
        urlbarSuggestions = object["urlbar-suggestions"];
        
        datalist = document.getElementById("tilepages-suggestions");
        
        for (i = 0; i < urlbarSuggestions.length; i++)
        {
            option = document.createElement("option");
            option.setAttribute("value",urlbarSuggestions[i]);
            datalist.appendChild(option);
        }
        
        /* Initialize navigation bar */
        
        navbarHeight = showNavBar ? 32 : 0;
        
        document.getElementById("tilepages-navbar").style.setProperty("display",showNavBar ? "flex" : "none","");
        document.getElementById("tilepages-container").style.setProperty("top",navbarHeight,"");
        
        /* Set view tiled button as checked */
        
        document.getElementById("tilepages-viewtiled").setAttribute("checked","");
        
        /* Add listeners */
        
        addListeners();
        
        /* Layout container ready */
        
        chrome.runtime.sendMessage({ type: "layoutReady" });
    });
}

/************************************************************************/

/* Add listeners */

function addListeners()
{
    /* Storage changed listener */
    
    chrome.storage.onChanged.addListener(
    function(changes,areaName)
    {
        chrome.storage.local.get(null,
        function(object)
        {
            showNavBar = object["options-shownavbar"];
            setTitle = object["options-settitle"];
            
            homePage = object["options-homepage"];
            
            syncActive = object["options-syncactive"];
            
            homePage = homePage.replace(/\/$/,"");
            
            if ("options-shownavbar" in changes)
            {
                toggleView();
                toggleView();
            }
        });
    });
    
    /* Add event listeners */
    
    window.addEventListener("resize",onResizeWindow,false);
    
    document.getElementById("tilepages-backward").addEventListener("click",onClickBackward,false);
    document.getElementById("tilepages-forward").addEventListener("click",onClickForward,false);
    document.getElementById("tilepages-reload").addEventListener("click",onClickReload,false);
    document.getElementById("tilepages-home").addEventListener("click",onClickHome,false);
    document.getElementById("tilepages-urlbar").addEventListener("focus",onFocusUrlBar,false);
    document.getElementById("tilepages-urlbar").addEventListener("blur",onBlurUrlBar,false);
    document.getElementById("tilepages-urlbar").addEventListener("drop",onDropUrlBar,false);
    document.getElementById("tilepages-urlbar").addEventListener("input",onInputUrlBar,false);
    document.getElementById("tilepages-urlbar").addEventListener("keypress",onKeypressUrlBar,false);
    document.getElementById("tilepages-go").addEventListener("click",onClickGo,false);
    document.getElementById("tilepages-bookmark").addEventListener("click",onClickBookmark,false);
    document.getElementById("tilepages-openintab").addEventListener("click",onClickOpenInTab,false);
    document.getElementById("tilepages-expandtile").addEventListener("click",onClickExpandTile,false);
    document.getElementById("tilepages-equalizetile").addEventListener("click",onClickEqualizeTile,false);
    document.getElementById("tilepages-untilepage").addEventListener("click",onClickUntilePage,false);
    document.getElementById("tilepages-syncscroll").addEventListener("click",onClickSyncScroll,false);
    document.getElementById("tilepages-viewtiled").addEventListener("click",onClickViewTiled,false);
    
    document.getElementById("tilepages-navbar").addEventListener("dragover",onDragOverNavBar,true);
    document.getElementById("tilepages-navbar").addEventListener("drop",onDropNavBar,true);

    document.getElementById("tilepages-container").addEventListener("mousedown",onMouseDownContainer,true);
    document.getElementById("tilepages-container").addEventListener("dblclick",onDblClickContainer,true);
    document.getElementById("tilepages-container").addEventListener("dragover",onDragOverContainer,true);
    document.getElementById("tilepages-container").addEventListener("dragleave",onDragLeaveContainer,true);
    document.getElementById("tilepages-container").addEventListener("drop",onDropContainer,true);
    
    /* Message received listener */
    
    chrome.runtime.onMessage.addListener(
    function(message,sender,sendResponse)
    {
        var tile,iframe;
        
        switch (message.type)
        {
            /* Messages from background page */
            
            case "newLayout":
                
                newLayout(message.layout,message.pageurls,message.tabid,message.incognito);
                
                break;
                
            case "tilePage":
                
                if (viewTiled) tilePage(activeTile,message.url,message.direction);
                
                break;
                
            case "tileDupPage":
                
                if (viewTiled) tileDupPage(activeTile,message.direction);
                
                break;
                
            case "assignPage":
                
                assignPage(activeTile,message.url);
                
                break;
                
            case "untilePage":
                
                if (viewTiled) untilePage(activeTile);
                
                break;
                
            case "expandTile":
                
                if (viewTiled) expandTile(activeTile);
                
                break;
                
            case "equalizeTile":
                
                if (viewTiled) equalizeTile(activeTile.parentTile);
                
                break;
                
            case "inspectTile":
                
                if (viewTiled) inspectTile(activeTile,message.node);
                
                break;
                
            case "toggleSync":
                
                toggleSync();
                
                break;
                
            case "toggleView":
                
                toggleView();
                
                break;
                
            case "openLayout":
                
                openLayout(message.layoutstring,message.layoutname,message.pageurls,message.tabid,message.incognito);
                
                break;
                
            case "saveLayout":
                
                saveLayout(message.layoutname,message.layouttype);
                
                break;
                
            /* Messages from content script loaded into web pages in iframes in container page */
            
            case "iframeLoad":
                
                onLoadIframe(message.panelid,message.url,message.title,sender.frameId);
                
                break;
                
            case "iframeMouseDown":
                
                onMouseDownIframe(message.panelid);
                
                break;
                
            case "iframeClick":
                
                onClickIframe(message.panelid,message.nextprev,message.shift,message.url);
                
                break;
                
            case "iframeDblClick":
                
                onDblClickIframe(message.panelid);
                
                break;
                
            case "iframeScroll":
                
                onScrollIframe(message.panelid,message.scrollx,message.scrolly);
                
                break;
                
            case "iframeDragStart":
                
                dragText = message.text;
                
                break;
                
            case "iframeDragOver":
                
                onDragOverIframe(message.panelid,message.shift,message.clientX,message.clientY);
                
                break;
                
            case "iframeDrop":
                
                onDropIframe(message.panelid,message.text,message.shift,message.clientX,message.clientY);
                
                break;
        }
    });
}

/************************************************************************/

/* Add tile functions */

function addTileFirst(url)
{
    /* Create tile object */
    
    rootTile = new Tile();
    rootTile.top = navbarHeight;
    rootTile.width = document.body.clientWidth;
    rootTile.height = document.body.clientHeight-navbarHeight;
    rootTile.panelId = createPanel(url);
    rootTile.historyIndex = 0;
    rootTile.historyUrl[0] = url;
    rootTile.historyTitle[0] = "";
    
    panelCount++;
    
    chrome.runtime.sendMessage({ type: "panelCount", count: panelCount });
    
    rootTile.scriptLoaded = false;
    document.getElementById(rootTile.panelId).style.setProperty("pointer-events","none","");
    
    return rootTile;
}

function addTile(tile,url,direction)
{
    var i,pc,newTile;
    
    /* If 'tile.type' is 'panel', then 'direction' can be any direction */
    /* If 'tile.type' is 'vsplit', then 'direction' must be 'left' or 'right' */
    /* If 'tile.type' is 'hsplit', then 'direction' must be 'above' or 'below' */
     
    if (tile.parentTile == null && tile.type != "panel")
    {
        rootTile = new Tile();
        rootTile.type = (direction == "left" || direction == "right") ? "vsplit" : "hsplit";
        rootTile.top = navbarHeight;
        rootTile.width = document.body.clientWidth;
        rootTile.height = document.body.clientHeight-navbarHeight;
        rootTile.childTiles[0] = tile;
        tile.parentTile = rootTile;
    }
    
    if (tile.parentTile != null &&
        ( (tile.parentTile.type == "vsplit" && (direction == "left" || direction == "right")) ||
          (tile.parentTile.type == "hsplit" && (direction == "above" || direction == "below")) ) )   /* add sibling panel tile to parent tile */
    {
        for (pc = 0; pc < tile.parentTile.childTiles.length; pc++) if (tile.parentTile.childTiles[pc] == tile) break;
        tile = tile.parentTile;
        
        switch (direction)
        {
            case "left":
                for (i = tile.childTiles.length-1; i >= pc; i--)
                {
                    tile.childTiles[i+1] = tile.childTiles[i];
                    if (i > pc) tile.splitterIds[i] = tile.splitterIds[i-1];
                }
                tile.childTiles[pc] = new Tile();
                tile.childTiles[pc].parentTile = tile;
                tile.childTiles[pc].width = Math.floor((tile.childTiles[pc+1].width-splitterSize)/2);
                tile.childTiles[pc].height = tile.height;
                tile.childTiles[pc].left = tile.childTiles[pc+1].left;
                tile.childTiles[pc].top = tile.top;
                tile.childTiles[pc].panelId = createPanel(url);
                tile.childTiles[pc].historyIndex = 0;
                tile.childTiles[pc].historyUrl[0] = url;
                tile.childTiles[pc].historyTitle[0] = "";
                tile.childTiles[pc+1].left = tile.childTiles[pc+1].left+splitterSize+Math.floor((tile.childTiles[pc+1].width-splitterSize)/2);
                tile.childTiles[pc+1].width = tile.childTiles[pc+1].width-splitterSize-Math.floor((tile.childTiles[pc+1].width-splitterSize)/2);
                tile.splitterIds[pc] = createSplitter();
                for (i = 0; i < tile.childTiles.length; i++)
                    tile.childTiles[i].ratio = tile.childTiles[i].width/(tile.width-tile.splitterIds.length*splitterSize);
                if (tile.childTiles[pc+1].type != "panel")
                    resizeTile(tile.childTiles[pc+1],tile.childTiles[pc+1].width,tile.height,tile.childTiles[pc+1].left,tile.top);
                break;
                
            case "right":
                for (i = tile.childTiles.length-1; i > pc; i--)
                {
                    tile.childTiles[i+1] = tile.childTiles[i];
                    tile.splitterIds[i] = tile.splitterIds[i-1];
                }
                tile.childTiles[pc+1] = new Tile();
                tile.childTiles[pc+1].parentTile = tile;
                tile.childTiles[pc+1].width = tile.childTiles[pc].width-splitterSize-Math.floor((tile.childTiles[pc].width-splitterSize)/2);
                tile.childTiles[pc+1].height = tile.height;
                tile.childTiles[pc+1].left = tile.childTiles[pc].left+splitterSize+Math.floor((tile.childTiles[pc].width-splitterSize)/2);
                tile.childTiles[pc+1].top = tile.top;
                tile.childTiles[pc+1].panelId = createPanel(url);
                tile.childTiles[pc+1].historyIndex = 0;
                tile.childTiles[pc+1].historyUrl[0] = url;
                tile.childTiles[pc+1].historyTitle[0] = "";
                tile.childTiles[pc].width = Math.floor((tile.childTiles[pc].width-splitterSize)/2);
                tile.splitterIds[pc] = createSplitter();
                for (i = 0; i < tile.childTiles.length; i++)
                    tile.childTiles[i].ratio = tile.childTiles[i].width/(tile.width-tile.splitterIds.length*splitterSize);
                if (tile.childTiles[pc].type != "panel")
                    resizeTile(tile.childTiles[pc],tile.childTiles[pc].width,tile.height,tile.childTiles[pc].left,tile.top);
                break;
                
            case "above":
                for (i = tile.childTiles.length-1; i >= pc; i--)
                {
                    tile.childTiles[i+1] = tile.childTiles[i];
                    if (i > pc) tile.splitterIds[i] = tile.splitterIds[i-1];
                }
                tile.childTiles[pc] = new Tile();
                tile.childTiles[pc].parentTile = tile;
                tile.childTiles[pc].width = tile.width;
                tile.childTiles[pc].height = Math.floor((tile.childTiles[pc+1].height-splitterSize)/2);
                tile.childTiles[pc].left = tile.left;
                tile.childTiles[pc].top = tile.childTiles[pc+1].top;
                tile.childTiles[pc].panelId = createPanel(url);
                tile.childTiles[pc].historyIndex = 0;
                tile.childTiles[pc].historyUrl[0] = url;
                tile.childTiles[pc].historyTitle[0] = "";
                tile.childTiles[pc+1].top = tile.childTiles[pc+1].top+splitterSize+Math.floor((tile.childTiles[pc+1].height-splitterSize)/2);
                tile.childTiles[pc+1].height = tile.childTiles[pc+1].height-splitterSize-Math.floor((tile.childTiles[pc+1].height-splitterSize)/2);
                tile.splitterIds[pc] = createSplitter();
                for (i = 0; i < tile.childTiles.length; i++)
                    tile.childTiles[i].ratio = tile.childTiles[i].height/(tile.height-tile.splitterIds.length*splitterSize);
                if (tile.childTiles[pc+1].type != "panel")
                    resizeTile(tile.childTiles[pc+1],tile.width,tile.childTiles[pc+1].height,tile.left,tile.childTiles[pc+1].top);
                break;
            
            case "below":
                for (i = tile.childTiles.length-1; i > pc; i--)
                {
                    tile.childTiles[i+1] = tile.childTiles[i];
                    tile.splitterIds[i] = tile.splitterIds[i-1];
                }
                tile.childTiles[pc+1] = new Tile();
                tile.childTiles[pc+1].parentTile = tile;
                tile.childTiles[pc+1].width = tile.width;
                tile.childTiles[pc+1].height = tile.childTiles[pc].height-splitterSize-Math.floor((tile.childTiles[pc].height-splitterSize)/2);
                tile.childTiles[pc+1].left = tile.left;
                tile.childTiles[pc+1].top = tile.childTiles[pc].top+splitterSize+Math.floor((tile.childTiles[pc].height-splitterSize)/2);
                tile.childTiles[pc+1].panelId = createPanel(url);
                tile.childTiles[pc+1].historyIndex = 0;
                tile.childTiles[pc+1].historyUrl[0] = url;
                tile.childTiles[pc+1].historyTitle[0] = "";
                tile.childTiles[pc].height = Math.floor((tile.childTiles[pc].height-splitterSize)/2);
                tile.splitterIds[pc] = createSplitter();
                for (i = 0; i < tile.childTiles.length; i++)
                    tile.childTiles[i].ratio = tile.childTiles[i].height/(tile.height-tile.splitterIds.length*splitterSize);
                if (tile.childTiles[pc].type != "panel")
                    resizeTile(tile.childTiles[pc],tile.width,tile.childTiles[pc].height,tile.left,tile.childTiles[pc].top);
                break;
        }
        
        newTile = (direction == "left" || direction == "above") ? tile.childTiles[pc] : tile.childTiles[pc+1];
    }
    else  /* split this panel tile into two child panel tiles */
    {
        tile.childTiles[0] = new Tile();
        tile.childTiles[0].parentTile = tile;
        tile.childTiles[0].width = (direction == "left" || direction == "right") ? Math.floor((tile.width-splitterSize)/2) : tile.width;
        tile.childTiles[0].height = (direction == "left" || direction == "right") ? tile.height : Math.floor((tile.height-splitterSize)/2);
        tile.childTiles[0].left = tile.left;
        tile.childTiles[0].top = tile.top;
        tile.childTiles[0].ratio = 1/2;
        tile.childTiles[0].panelId = (direction == "left" || direction == "above") ? createPanel(url) : tile.panelId;
        tile.childTiles[0].scrollX = (direction == "left" || direction == "above") ? 0 : tile.scrollX;  /* 0 until getScrollPosn updates */
        tile.childTiles[0].scrollY = (direction == "left" || direction == "above") ? 0 : tile.scrollY;  /* 0 until getScrollPosn updates */
        tile.childTiles[0].historyIndex = (direction == "left" || direction == "above") ? 0 : tile.historyIndex;
        tile.childTiles[0].historyUrl = (direction == "left" || direction == "above") ? new Array(url) : tile.historyUrl.slice(0);
        tile.childTiles[0].historyTitle = (direction == "left" || direction == "above") ? new Array() : tile.historyTitle.slice(0);
        tile.childTiles[1] = new Tile();
        tile.childTiles[1].parentTile = tile;
        tile.childTiles[1].width = (direction == "left" || direction == "right") ? tile.width-splitterSize-tile.childTiles[0].width : tile.width;
        tile.childTiles[1].height = (direction == "left" || direction == "right") ? tile.height : tile.height-splitterSize-tile.childTiles[0].height;
        tile.childTiles[1].left = (direction == "left" || direction == "right") ? tile.left+splitterSize+tile.childTiles[0].width : tile.left;
        tile.childTiles[1].top = (direction == "left" || direction == "right") ? tile.top : tile.top+splitterSize+tile.childTiles[0].height;
        tile.childTiles[1].ratio = 1/2;
        tile.childTiles[1].panelId = (direction == "left" || direction == "above") ? tile.panelId : createPanel(url);
        tile.childTiles[1].scrollX = (direction == "left" || direction == "above") ? tile.scrollX : 0;  /* 0 until getScrollPosn updates */
        tile.childTiles[1].scrollY = (direction == "left" || direction == "above") ? tile.scrollY : 0;  /* 0 until getScrollPosn updates */
        tile.childTiles[1].historyIndex = (direction == "left" || direction == "above") ? tile.historyIndex : 0;
        tile.childTiles[1].historyUrl = (direction == "left" || direction == "above") ? tile.historyUrl.slice(0) : new Array(url);
        tile.childTiles[1].historyTitle = (direction == "left" || direction == "above") ? tile.historyTitle.slice(0) : new Array();
        tile.type = (direction == "left" || direction == "right") ? "vsplit" : "hsplit";
        tile.panelId = "";
        tile.splitterIds[0] = createSplitter();
        
        newTile = (direction == "left" || direction == "above") ? tile.childTiles[0] : tile.childTiles[1];
    }
    
    panelCount++;
    
    chrome.runtime.sendMessage({ type: "panelCount", count: panelCount });
    
    newTile.scriptLoaded = false;
    document.getElementById(newTile.panelId).style.setProperty("pointer-events","none","");
    
    return newTile;
}

/************************************************************************/
    
/* Remove tile functions */

function removeTile(tile)
{
    var i,num,pc,gpc,gptile,ptile,stile,actTile;
    
    ptile = tile.parentTile;
    for (pc = 0; pc < ptile.childTiles.length; pc++) if (ptile.childTiles[pc] == tile) break;
    if (pc == 0) stile = ptile.childTiles[1]; else stile = ptile.childTiles[pc-1];  /* sibling tile */
    
    if (pc == 0) destroySplitter(ptile.splitterIds[0]);
    else destroySplitter(ptile.splitterIds[pc-1]);
    
    destroyPanel(tile.panelId);
    
    if (ptile.childTiles.length == 2)  /* exactly 2 children */
    {
        if (stile.type == "panel")  /* 2 panel tiles */
        {
            ptile.type = "panel";
            ptile.panelId = stile.panelId;
            ptile.scrollX = stile.scrollX;
            ptile.scrollY = stile.scrollY;
            ptile.historyIndex = stile.historyIndex;
            ptile.historyUrl = stile.historyUrl.slice(0);
            ptile.historyTitle = stile.historyTitle.slice(0);
            
            ptile.childTiles.length = 0;
            ptile.splitterIds.length = 0;
            
            panelCount--;
            
            chrome.runtime.sendMessage({ type: "panelCount", count: panelCount });
            
            return stile;
        }
        else  /* 1 panel tile and 1 hsplit/vsplit tile */
        {
            for (actTile = stile.childTiles[0]; actTile.type != "panel"; actTile = actTile.childTiles[0]);
            
            if (ptile.parentTile == null) /* at root */
            {
                if (ptile.type == "hsplit")  /* horizontal split */
                {
                    stile.height = ptile.height;
                    stile.top = ptile.top;
                    stile.ratio = 1;
                    stile.parentTile = null;
                }
                else  /* vertical split */
                {
                    stile.width = ptile.width;
                    stile.left = ptile.left;
                    stile.ratio = 1;
                    stile.parentTile = null;
                }
                
                ptile.childTiles.length = 0;
                ptile.splitterIds.length = 0;
                
                resizeTile(stile,stile.width,stile.height,stile.left,stile.top);
                
                rootTile = stile;
            }
            else  /* not at root */
            {
                num = stile.childTiles.length;
                gptile = ptile.parentTile;
                for (gpc = 0; gpc < gptile.childTiles.length; gpc++) if (gptile.childTiles[gpc] == ptile) break;
                for (i = gptile.childTiles.length-1; i > gpc; i--)
                {
                    gptile.childTiles[i+num-1] = gptile.childTiles[i];
                    gptile.splitterIds[i+num-2] = gptile.splitterIds[i-1];
                }
                for (i = 0; i < num; i++) gptile.childTiles[gpc+i] = stile.childTiles[i];
                for (i = 0; i < num-1; i++) gptile.splitterIds[gpc+i] = stile.splitterIds[i];
                if (gptile.type == "hsplit")  /* horizontal split */
                {
                    for (i = 0; i < num; i++)
                    {
                        gptile.childTiles[gpc+i].parentTile = gptile;
                        gptile.childTiles[gpc+i].width = gptile.width;
                        gptile.childTiles[gpc+i].left = gptile.left;
                    }
                    for (i = 0; i < gptile.childTiles.length; i++)
                      gptile.childTiles[i].ratio = gptile.childTiles[i].height/(gptile.height-gptile.splitterIds.length*splitterSize);
                }
                else  /* vertical split */
                {
                    for (i = 0; i < num; i++)
                    {
                        gptile.childTiles[gpc+i].parentTile = gptile;
                        gptile.childTiles[gpc+i].height = gptile.height;
                        gptile.childTiles[gpc+i].top = gptile.top;
                    }
                    for (i = 0; i < gptile.childTiles.length; i++)
                        gptile.childTiles[i].ratio = gptile.childTiles[i].width/(gptile.width-gptile.splitterIds.length*splitterSize);
                }
                
                stile.childTiles.length = 0;
                stile.splitterIds.length = 0;
                ptile.childTiles.length = 0;
                ptile.splitterIds.length = 0;
                
                resizeTile(gptile,gptile.width,gptile.height,gptile.left,gptile.top);
            }
            
            panelCount--;
            
            chrome.runtime.sendMessage({ type: "panelCount", count: panelCount });
            
            return actTile;
        }
    }
    else  /* more than 2 children */
    {
        for (actTile = stile; actTile.type != "panel"; actTile = actTile.childTiles[0]);
        
        if (ptile.type == "hsplit")  /* horizontal split */
        {
            stile.height += tile.height+splitterSize;
            if (pc == 0) stile.top = tile.top;
            for (i = 0; i < ptile.childTiles.length; i++) if (i != pc)
                ptile.childTiles[i].ratio = ptile.childTiles[i].height/(ptile.height-(ptile.splitterIds.length-1)*splitterSize);
        }
        else  /* vertical split */
        {
            stile.width += tile.width+splitterSize;
            if (pc == 0) stile.left = tile.left;
            for (i = 0; i < ptile.childTiles.length; i++) if (i != pc)
                ptile.childTiles[i].ratio = ptile.childTiles[i].width/(ptile.width-(ptile.splitterIds.length-1)*splitterSize);
        }
        for (i = pc; i < ptile.childTiles.length-1; i++)
        {
            ptile.childTiles[i] = ptile.childTiles[i+1];
            if (i > 0) ptile.splitterIds[i-1] = ptile.splitterIds[i];
        }
        
        ptile.childTiles.length--;
        ptile.splitterIds.length--;
        
        resizeTile(ptile,ptile.width,ptile.height,ptile.left,ptile.top);
        
        panelCount--;
        
        chrome.runtime.sendMessage({ type: "panelCount", count: panelCount });
        
        return actTile;
    }
}

function removeTileLast()
{
    destroyPanel(rootTile.panelId);
    
    rootTile = null;
    
    panelCount--;
    
    chrome.runtime.sendMessage({ type: "panelCount", count: panelCount });
    
    return null;
}

/************************************************************************/

/* Resize tile functions */

function balanceTile(tile,nested)
{
    var i;
    
    if (tile.type == "panel")  /* panel tile */
    {
    }
    else  /* split tile */
    {
        for (i = 0; i < tile.childTiles.length; i++)
        {
            tile.childTiles[i].ratio = 1/tile.childTiles.length;
            if (nested) balanceTile(tile.childTiles[i],nested);
        }
    }
}

function enlargeTile(tile)
{
    var i,enlargetile;
    
    enlargetile = tile;
    
    while (tile.parentTile != null && tile.ratio == 1)
    {
        tile = tile.parentTile;
    }
    
    if (tile.parentTile != null)
    {
        tile.expandRatio = tile.ratio;
        tile.ratio = 1;
        
        for (i = 0; i < tile.parentTile.childTiles.length; i++)
        {
            if (tile.parentTile.childTiles[i] != tile)
            {
                tile.parentTile.childTiles[i].expandRatio = tile.parentTile.childTiles[i].ratio;
                tile.parentTile.childTiles[i].ratio = 0;
            }
        }
    }
    else
    {
        tile = enlargetile;
        
        while (tile.parentTile != null)
        {
            for (i = 0; i < tile.parentTile.childTiles.length; i++) tile.parentTile.childTiles[i].ratio = tile.parentTile.childTiles[i].expandRatio;
            
            tile = tile.parentTile;
        }
    }
}

function resizeTile(tile,newWidth,newHeight,newLeft,newTop)
{
    var i,width,height,left,top;
   
    if (tile.type == "panel")  /* panel tile */
    {
    }
    else  /* split tile */
    {
        if (tile.type == "hsplit")  /* horizontal split */
        {
            if (newHeight >= tile.splitterIds.length*splitterSize)
            {
                top = newTop;
                for (i = 0; i < tile.childTiles.length-1; i++)
                {
                    height = Math.floor((newHeight-tile.splitterIds.length*splitterSize)*tile.childTiles[i].ratio);
                    resizeTile(tile.childTiles[i],newWidth,height,newLeft,top);
                    top += tile.childTiles[i].height+splitterSize;
                }
                resizeTile(tile.childTiles[i],newWidth,newHeight-(top-newTop),newLeft,top);
            }
            else
            {
                for (i = 0; i < tile.childTiles.length; i++)
                {
                    resizeTile(tile.childTiles[i],newWidth,0,newLeft,newTop);
                }
            }
        }
        else  /* vertical split */
        {
            if (newWidth >= tile.splitterIds.length*splitterSize)
            {
                left = newLeft;
                for (i = 0; i < tile.childTiles.length-1; i++)
                {
                    width = Math.floor((newWidth-tile.splitterIds.length*splitterSize)*tile.childTiles[i].ratio);
                    resizeTile(tile.childTiles[i],width,newHeight,left,newTop);
                    left += tile.childTiles[i].width+splitterSize;
                }
                resizeTile(tile.childTiles[i],newWidth-(left-newLeft),newHeight,left,newTop);
            }
            else
            {
                for (i = 0; i < tile.childTiles.length; i++)
                {
                    resizeTile(tile.childTiles[i],0,newHeight,newLeft,newTop);
                }
            }
        }
    }
    
    tile.width = newWidth;
    tile.height = newHeight;
    tile.left = newLeft;
    tile.top = newTop;
}

/************************************************************************/

/* Create panel/splitter functions */

function createPanel(url)
{
    var iframe;
    
    /* Create <iframe> element for panel */
    
    iframe = document.createElement("iframe");
    iframe.setAttribute("id","tilepages-panel-" + panelSeqNo);
    iframe.setAttribute("name","tilepages-panel-" + panelSeqNo++);
    iframe.setAttribute("src",url);
    
    // iframe.setAttribute("sandbox","");
    
    // iframe.setAttribute("sandbox","allow-forms allow-modals allow-orientation-lock allow-pointer-lock allow-popups allow-popups-to-escape-sandbox allow-presentation allow-scripts allow-storage-access-by-user-activation allow-top-navigation allow-top-navigation-by-user-activation allow-same-origin");  /*???*/
    
    // iframe.setAttribute("sandbox","allow-forms");  /*???*/
    // iframe.setAttribute("sandbox","allow-modals");  /*???*/
    // iframe.setAttribute("sandbox","allow-orientation-lock");  /*???*/
    // iframe.setAttribute("sandbox","allow-pointer-lock");  /*???*/
    // iframe.setAttribute("sandbox","allow-popups");  /*???*/
    // iframe.setAttribute("sandbox","allow-popups-to-escape-sandbox");  /*???*/
    // iframe.setAttribute("sandbox","allow-presentation");  /*???*/
    // iframe.setAttribute("sandbox","allow-same-origin");  /*???*/
    // iframe.setAttribute("sandbox","allow-scripts");  /*???*/
    // iframe.setAttribute("sandbox","allow-storage-access-by-user-activation");  /*???*/
    // iframe.setAttribute("sandbox","allow-top-navigation");  /*???*/
    // iframe.setAttribute("sandbox","allow-top-navigation-by-user-activation");  /*???*/
    
    iframe.setAttribute("sandbox","allow-forms allow-same-origin");  /*???*/  /* Firefox - allow-same-origin necessary for content scripts to execute */
    
    iframe.style.setProperty("position","fixed","");
    iframe.style.setProperty("box-sizing","border-box","");
    iframe.style.setProperty("z-index","0","");
    document.getElementById("tilepages-container").appendChild(iframe);
    
    return iframe.getAttribute("id");
}

function destroyPanel(panelId)
{
    var iframe;
    
    iframe = document.getElementById(panelId);
    
    document.getElementById("tilepages-container").removeChild(iframe);
}

function createSplitter()
{
    var div;
    
    /* Create <div> element for splitter */
    
    div = document.createElement("div");
    div.setAttribute("id","tilepages-splitter-" + splitterSeqNo++);
    div.style.setProperty("position","fixed","");
    div.style.setProperty("box-sizing","border-box","");
    div.style.setProperty("z-index","2","");
    document.getElementById("tilepages-container").appendChild(div);
    
    /* Add listeners for splitter */
    
    div.addEventListener("dblclick",onDblClickSplitter,false);
    div.addEventListener("mousedown",onMouseDownSplitter,false);
    
    return div.getAttribute("id");
}

function destroySplitter(splitterId)
{
    var div;
    
    div = document.getElementById(splitterId);
    
    /* Remove listeners for splitter */
    
    div.removeEventListener("dblclick",onDblClickSplitter,false);
    div.removeEventListener("mousedown",onMouseDownSplitter,false);
    
    /* Destroy <div> element for splitter */
    
    document.getElementById("tilepages-container").removeChild(div);
}

/************************************************************************/

/* Draw tile function */

function drawTile(tile,show)
{
    var i,left,top,iframe,div;
    
    if (tile.type == "panel")  /* panel tile */
    {
        iframe = document.getElementById(tile.panelId);
        
        iframe.style.setProperty("display",show ? "block" : "none","");
        iframe.style.setProperty("left",tile.left + "px","");
        iframe.style.setProperty("top",tile.top + "px","");
        iframe.style.setProperty("width",tile.width + "px","");
        iframe.style.setProperty("height",tile.height + "px","");
        iframe.style.setProperty("border",tileBorder + "px solid transparent","");
    }
    else  /* split tile */
    {
        if (tile.type == "hsplit")  /* horizontal split */
        {
            if (tile.height >= tile.splitterIds.length*splitterSize)
            {
                top = tile.top-splitterSize;
                
                for (i = 0; i < tile.splitterIds.length; i++)
                {
                    top += tile.childTiles[i].height+splitterSize;
                    
                    div = document.getElementById(tile.splitterIds[i]);
                    
                    div.style.setProperty("display",show ? "block" : "none","");
                    div.style.setProperty("left",tile.left + "px","");
                    div.style.setProperty("top",top + "px","");
                    div.style.setProperty("width",tile.width + "px","");
                    div.style.setProperty("height",splitterSize + "px","");
                    div.style.setProperty("border-top","1px solid  #A0A0A0","");
                    div.style.setProperty("border-bottom","1px solid  #A0A0A0","");
                    div.style.setProperty("background","#E1E6EF","");
                    div.style.setProperty("cursor","url(" + chrome.runtime.getURL("img-cursor-ns.png") + ") 4 11,ns-resize","");
                }
            }
            else
            {
                for (i = 0; i < tile.splitterIds.length; i++)
                {
                    div = document.getElementById(tile.splitterIds[i]);
                    div.style.setProperty("visibility","hidden","");
                }
            }
        }
        else  /* vertical split */
        {
            if (tile.width >= tile.splitterIds.length*splitterSize)
            {
                left = tile.left-splitterSize;
                
                for (i = 0; i < tile.splitterIds.length; i++)
                {
                    left += tile.childTiles[i].width+splitterSize;
                    
                    div = document.getElementById(tile.splitterIds[i]);
                    
                    div.style.setProperty("display",show ? "block" : "none","");
                    div.style.setProperty("left",left + "px","");
                    div.style.setProperty("top",tile.top + "px","");
                    div.style.setProperty("width",splitterSize + "px","");
                    div.style.setProperty("height",tile.height + "px","");
                    div.style.setProperty("border-left","1px solid  #A0A0A0","");
                    div.style.setProperty("border-right","1px solid  #A0A0A0","");
                    div.style.setProperty("background","#E1E6EF","");
                    div.style.setProperty("cursor","url(" + chrome.runtime.getURL("img-cursor-ew.png") + ") 11 4,ew-resize","");
                }
            }
            else
            {
                for (i = 0; i < tile.splitterIds.length; i++)
                {
                    div = document.getElementById(tile.splitterIds[i]);
                    div.style.setProperty("visibility","hidden","");
                }
            }
        }
        
        for (i = 0; i < tile.childTiles.length; i++)
            drawTile(tile.childTiles[i],show);
    }
}

/************************************************************************/

/* Toggle Sync */

function toggleSync()
{
    var iframe,div;
    
    syncScroll = !syncScroll;
    
    chrome.runtime.sendMessage({ type: "syncScroll", state: syncScroll });
    
    if (!syncScroll)
    {
        document.getElementById("tilepages-syncscroll").removeAttribute("checked","");
    }
    else
    {
        document.getElementById("tilepages-syncscroll").setAttribute("checked","");
    }
}

/************************************************************************/

/* Toggle View */

function toggleView()
{
    var iframe,div;
    
    viewTiled = !viewTiled;
    
    chrome.runtime.sendMessage({ type: "viewTiled", state: viewTiled });
    
    if (!viewTiled)
    {
        document.getElementById("tilepages-viewtiled").removeAttribute("checked","");
        
        document.getElementById("tilepages-expandtile").disabled = true;
        document.getElementById("tilepages-equalizetile").disabled = true;
        document.getElementById("tilepages-untilepage").disabled = true;
        
        navbarHeight = 32;
        
        document.getElementById("tilepages-navbar").style.setProperty("display","flex","");
        document.getElementById("tilepages-container").style.setProperty("top",navbarHeight + "px","");
        
        resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
        
        drawTile(rootTile,false);
        
        iframe = document.getElementById(activeTile.panelId);
        
        iframe.style.setProperty("display","block","");
        iframe.style.setProperty("left","0px","");
        iframe.style.setProperty("top",navbarHeight + "px","");
        iframe.style.setProperty("width",document.body.clientWidth + "px","");
        iframe.style.setProperty("height",(document.body.clientHeight-navbarHeight) + "px","");
        iframe.style.setProperty("border","none","");
    }
    else
    {
        document.getElementById("tilepages-viewtiled").setAttribute("checked","");
        
        document.getElementById("tilepages-expandtile").disabled = false;
        document.getElementById("tilepages-equalizetile").disabled = false;
        document.getElementById("tilepages-untilepage").disabled = false;
        
        navbarHeight = showNavBar ? 32 : 0;
        
        resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
        
        document.getElementById("tilepages-navbar").style.setProperty("display",showNavBar ? "flex" : "none","");
        document.getElementById("tilepages-container").style.setProperty("top",navbarHeight + "px","");
        
        drawTile(rootTile,true);
        
        iframe = document.getElementById(activeTile.panelId);
        
        iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
    }
}

/************************************************************************/

/* New Layout */

function newLayout(layout,pageUrls,tabId,incognito)
{
    var i,j,cols,rows,tile,iframe,urlbar,titlebar,button;
    
    layoutTabId = tabId;
    
    layoutTabIncognito = incognito;
    
    switch (layout)
    {
        case "2vert":
            
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"right");
            
            break;
            
        case "2horiz":
            
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"below");
            
            break;
            
        case "3vert":
            
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"right");
            addTile(rootTile.childTiles[1],allocateUrl(pageUrls,2),"right");
            balanceTile(rootTile,false);
            resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
            
            break;
            
        case "3horiz":
            
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"below");
            addTile(rootTile.childTiles[1],allocateUrl(pageUrls,2),"below");
            balanceTile(rootTile,false);
            resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
            
            break;
            
        case "4vertgrid":
            
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"below");
            addTile(rootTile,allocateUrl(pageUrls,2),"right");
            addTile(rootTile.childTiles[1],allocateUrl(pageUrls,3),"below");
            
            break;
            
        case "4horizgrid":
            
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"right");
            addTile(rootTile,allocateUrl(pageUrls,2),"below");
            addTile(rootTile.childTiles[1],allocateUrl(pageUrls,3),"right");
            
            break;
            
        case "6vertgrid":
            
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"below");
            addTile(rootTile,allocateUrl(pageUrls,2),"right");
            addTile(rootTile.childTiles[1],allocateUrl(pageUrls,3),"below");
            addTile(rootTile.childTiles[1],allocateUrl(pageUrls,4),"right");
            addTile(rootTile.childTiles[2],allocateUrl(pageUrls,5),"below");
            balanceTile(rootTile,true);
            resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
            
            break;
            
        case "6horizgrid":
            
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"right");
            addTile(rootTile.childTiles[1],allocateUrl(pageUrls,2),"right");
            addTile(rootTile,allocateUrl(pageUrls,3),"below");
            addTile(rootTile.childTiles[1],allocateUrl(pageUrls,4),"right");
            addTile(rootTile.childTiles[1].childTiles[1],allocateUrl(pageUrls,5),"right");
            balanceTile(rootTile,true);
            resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
            
            break;
            
        case "allvert":
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"right");
            for (i = 2; i < pageUrls.length; i++) addTile(rootTile.childTiles[i-1],allocateUrl(pageUrls,i),"right");
            balanceTile(rootTile,false);
            resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
            break;
            
        case "allhoriz":
            addTileFirst(allocateUrl(pageUrls,0));
            addTile(rootTile,allocateUrl(pageUrls,1),"below");
            for (i = 2; i < pageUrls.length; i++) addTile(rootTile.childTiles[i-1],allocateUrl(pageUrls,i),"below");
            balanceTile(rootTile,false);
            resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
            break;
            
        case "allvertgrid":
            
            tile = addTileFirst(allocateUrl(pageUrls,0));
            cols = Math.ceil(Math.sqrt(pageUrls.length));
            rows = Math.ceil(pageUrls.length/cols);
            for (i = 1; i < cols; i++) tile = addTile(tile,allocateUrl(pageUrls,i*rows),"right");
            for (i = 0; i < cols; i++)
            {
                tile = rootTile.childTiles[i];
                for (j = 1; j < rows; j++) tile = addTile(tile,allocateUrl(pageUrls,i*rows+j),"below");
            }
            balanceTile(rootTile,true);
            resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
            
            break;
            
        case "allhorizgrid":
            
            tile = addTileFirst(allocateUrl(pageUrls,0));
            rows = Math.ceil(Math.sqrt(pageUrls.length));
            cols = Math.ceil(pageUrls.length/rows);
            for (i = 1; i < rows; i++) tile = addTile(tile,allocateUrl(pageUrls,i*cols),"below");
            for (i = 0; i < rows; i++)
            {
                tile = rootTile.childTiles[i];
                for (j = 1; j < cols; j++) tile = addTile(tile,allocateUrl(pageUrls,i*cols+j),"right");
            }
            balanceTile(rootTile,true);
            resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
            
            break;
    }
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    drawTile(rootTile,true);
    
    for (tile = rootTile; tile.type != "panel"; tile = tile.childTiles[0]);
    
    activeTile = tile;
    
    focusActiveTile();
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = activeTile.historyUrl[activeTile.historyIndex];
    
    titlebar = document.getElementById("tilepages-titlebar");
    titlebar.value = "";
    
    button = document.getElementById("tilepages-go");
    button.disabled = true;
    
    button = document.getElementById("tilepages-backward");
    button.disabled = true;
    
    button = document.getElementById("tilepages-forward");
    button.disabled = true;
}

/************************************************************************/

/* Tile Pages */

function tilePage(tile,url,direction)
{
    var iframe,urlbar,titlebar,button;
    
    activeTile = addTile(tile,url,direction);
    
    focusActiveTile();
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    drawTile(rootTile,true);
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = url;
    
    titlebar = document.getElementById("tilepages-titlebar");
    titlebar.value = "";
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
    
    button = document.getElementById("tilepages-backward");
    button.disabled = true;
    
    button = document.getElementById("tilepages-forward");
    button.disabled = true;
}

function tileDupPage(tile,direction)
{
    var iframe,urlbar,titlebar,button;
    
    activeTile = addTile(tile,tile.historyUrl[tile.historyIndex],direction);
    
    focusActiveTile();
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    drawTile(rootTile,true);
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = url;
    
    titlebar = document.getElementById("tilepages-titlebar");
    titlebar.value = "";
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
    
    button = document.getElementById("tilepages-backward");
    button.disabled = true;
    
    button = document.getElementById("tilepages-forward");
    button.disabled = true;
}

/************************************************************************/

/* Assign Page */

function assignPage(tile,url)
{
    var urlbar,titlebar,iframe,button;
    
    urlbar = document.getElementById("tilepages-urlbar");
    titlebar = document.getElementById("tilepages-titlebar");
    
    if (tile == activeTile)
    {
        urlbar.value = url;
        titlebar.value = "";
    }
    
    iframe = document.getElementById(tile.panelId);
    iframe.src = url;
    
    tile.scriptLoaded = false;
    iframe.style.setProperty("pointer-events","none","");
    
    if (url != tile.historyUrl[tile.historyIndex])
    {
        tile.historyIndex++;
        tile.historyUrl.length = tile.historyIndex;
        tile.historyUrl[tile.historyIndex] = url;
        tile.historyTitle.length = tile.historyIndex;
        tile.historyTitle[tile.historyIndex] = "";
    }
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
    
    button = document.getElementById("tilepages-backward");
    button.disabled = (activeTile.historyIndex <= 0);
    
    button = document.getElementById("tilepages-forward");
    button.disabled = (activeTile.historyIndex >= activeTile.historyUrl.length-1);
}

/************************************************************************/

/* Untile Page */

function untilePage(tile)
{
    var iframe,urlbar,titlebar,button;
    
    if (panelCount <= 2)
    {
        chrome.runtime.sendMessage({ type: "closeLayout" });
        
        return;
    }
    
    activeTile = removeTile(tile);
    
    focusActiveTile();
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    drawTile(rootTile,true);
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = activeTile.historyTitle[activeTile.historyIndex];
    
    titlebar = document.getElementById("tilepages-titlebar");
    titlebar.value = activeTile.historyTitle[activeTile.historyIndex];
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
    
    button = document.getElementById("tilepages-backward");
    button.disabled = (activeTile.historyIndex <= 0);
    
    button = document.getElementById("tilepages-forward");
    button.disabled = (activeTile.historyIndex >= activeTile.historyUrl.length-1);
}

/************************************************************************/

/* Expand Tile */

function expandTile(tile)
{
    var iframe;
    
    enlargeTile(tile);
    
    resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
    
    saveRestoreLayout();
    
    drawTile(rootTile,true);
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
}

/************************************************************************/

/* Equalize Tile */

function equalizeTile(tile)
{
    var iframe;
    
    balanceTile(tile,false);
    
    resizeTile(tile,tile.width,tile.height,tile.left,tile.top);
    
    saveRestoreLayout();
    
    drawTile(tile,true);
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
}

/************************************************************************/

/* Inspect Tile */

function inspectTile(tile,node)
{
    var i;
    
    i = tile.parentTile.childTiles.indexOf(tile);
    
    if (node == "this") tile = tile;
    else if (node == "before" && i > 0) tile = tile.parentTile.childTiles[i-1];
    else if (node == "after" && i < tile.parentTile.childTiles.length-1) tile = tile.parentTile.childTiles[i+1];
    else if (node == "parent" && tile.parentTile != null) tile = tile.parentTile;
    else if (node == "gparent" && tile.parentTile != null && tile.parentTile.parentTile != null) tile = tile.parentTile.parentTile;
    else if (node == "ggparent" && tile.parentTile != null && tile.parentTile.parentTile != null && tile.parentTile.parentTile.parentTile != null) tile = tile.parentTile.parentTile.parentTile;
    else { alert("Tile does not exist"); return; }
    
    alert(tile.type + " >  ID: " + activeTile.panelId + "  L: " + tile.left + "  T: " + tile.top + "  W: " + tile.width + "  H: " + tile.height + "  R: " + tile.ratio.toFixed(4));
}

/************************************************************************/

/* Open Layout */

function openLayout(layoutString,layoutName,pageUrls,tabId,incognito)
{
    var tile,iframe,urlbar,titlebar,button;
    
    layoutTabId = tabId;
    
    layoutTabIncognito = incognito;
    
    if (setTitle) document.title = layoutName.replace(/~/g," ");
    
    applyLayoutString(layoutString,pageUrls);
    
    resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    drawTile(rootTile,true);
    
    for (tile = rootTile; tile.type != "panel"; tile = tile.childTiles[0]);
    
    activeTile = tile;
    
    focusActiveTile();
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = activeTile.historyUrl[activeTile.historyIndex];
    
    titlebar = document.getElementById("tilepages-titlebar");
    titlebar.value = "";
    
    button = document.getElementById("tilepages-go");
    button.disabled = true;
    
    button = document.getElementById("tilepages-backward");
    button.disabled = true;
    
    button = document.getElementById("tilepages-forward");
    button.disabled = true;
}

function applyLayoutString(layoutString,pageUrls)
{
    var j,layoutType;
    
    j = layoutString.indexOf(",");
    layoutType = +layoutString.substr(0,j);
    layoutString = layoutString.substr(j+1);
    
    j = layoutString.indexOf(";");
    panelCount = +layoutString.substr(0,j);
    layoutString = layoutString.substr(j+1);
    
    panelCount = 0;
    
    chrome.runtime.sendMessage({ type: "panelCount", count: panelCount });

    rootTile = new Tile();
    
    applyTileString(rootTile);
    
    function applyTileString(tile)
    {
        var i,j,url,childCount;
        
        j = layoutString.indexOf(",");
        tile.type = layoutString.substr(0,j);
        layoutString = layoutString.substr(j+1);
        
        if (tile.type == "panel")  /* panel tile */
        {
            if (layoutType == 0)
            {
                j = layoutString.indexOf(";");
                tile.ratio = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                url = allocateUrl(pageUrls,panelCount);
            }
            else if (layoutType == 1)
            {
                j = layoutString.indexOf(",");
                tile.ratio = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over { */
                
                j = layoutString.indexOf("}");
                url = layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over } */
            }
            else if (layoutType == 2)
            {
                j = layoutString.indexOf(",");
                tile.ratio = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over { */
                
                j = layoutString.indexOf("}");
                url = layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over } */
                
                j = layoutString.indexOf(",");
                tile.scrollX = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                j = layoutString.indexOf(";");
                tile.scrollY = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                tile.scrollInit = true;
            }
                
            if (restrictedPage(url)) url = (homePage != "") ? homePage : "about:blank";
            
            tile.panelId = createPanel(url);
            
            tile.scriptLoaded = false;
            document.getElementById(tile.panelId).style.setProperty("pointer-events","none","");
            
            tile.historyIndex = 0;
            tile.historyUrl[0] = url;
            tile.historyTitle[0] = "";
            
            panelCount++;
            
            chrome.runtime.sendMessage({ type: "panelCount", count: panelCount });
        }
        else  /* split tile */
        {
            j = layoutString.indexOf(",");
            tile.ratio = +layoutString.substr(0,j);
            layoutString = layoutString.substr(j+1);
            
            j = layoutString.indexOf(";");
            childCount = +layoutString.substr(0,j);
            layoutString = layoutString.substr(j+1);
            
            for (i = 0; i < childCount; i++)
            {
                tile.childTiles[i] = new Tile();
                
                tile.childTiles[i].parentTile = tile;
                
                if ( i > 0) tile.splitterIds[i-1] = createSplitter();
                
                applyTileString(tile.childTiles[i]);
            }
        }
    }
}

/************************************************************************/

/* Save Layout */

function saveLayout(layoutName,layoutType)
{
    var object = new Object();
    
    if (layoutName == "")  object["layout-restore"] = createLayoutString(rootTile,layoutType);
    else object["layout-string-" + layoutName] = createLayoutString(rootTile,layoutType);
    
    chrome.storage.local.set(object);
}

function createLayoutString(tile,layoutType)
{
    var layoutString;
    
    layoutString = layoutType + "," + panelCount + ";";
    
    createTileString(tile);
    
    return layoutString;
    
    function createTileString(tile)
    {
        var i,iframe;
        
        if (tile.type == "panel")
        {
            iframe = document.getElementById(tile.panelId);
            
            if (layoutType == 0)
            {
                layoutString += "panel" + "," + tile.ratio.toFixed(4) + ";";
            }
            else if (layoutType == 1)
            {
                layoutString += "panel" + "," + tile.ratio.toFixed(4) + "," + "{" + tile.historyUrl[tile.historyIndex] + "}" + ";";
            }
            else if (layoutType == 2)
            {
                layoutString += "panel" + "," + tile.ratio.toFixed(4) + "," + "{" + tile.historyUrl[tile.historyIndex] + "}" + "," + tile.scrollX + "," + tile.scrollY + ";";
            }
        }
        else  /* split tile */
        {
            layoutString += tile.type + "," + tile.ratio.toFixed(4) + "," + tile.childTiles.length + ";";
            
            for (i = 0; i < tile.childTiles.length; i++)
                createTileString(tile.childTiles[i],layoutType);
        }
    }
}

/************************************************************************/

/* Resize window event function */

function onResizeWindow(event)
{
    var iframe;
    
    resizeTile(rootTile,document.body.clientWidth,document.body.clientHeight-navbarHeight,0,navbarHeight);
    
    saveRestoreLayout();
    
    drawTile(rootTile,true);
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
}

/************************************************************************/

/* Splitter event functions */

function onMouseDownSplitter(event)
{
    var i,iframes,iframe,container;
    
    lastScreenX = event.screenX;
    lastScreenY = event.screenY;
    
    downTile = findTileBySplitter(rootTile,event.currentTarget.id);
    
    for (i = 0; i < downTile.splitterIds.length; i++) if (downTile.splitterIds[i] == event.currentTarget.id) break;
    downSplitter = i;
    
    iframes = document.querySelectorAll("[name^=tilepages-panel-]");
    
    for (i = 0; i < iframes.length; i++)
    {
        iframe = document.getElementById(iframes[i].name);
        iframe.style.setProperty("pointer-events","none","");
    }
    
    container = document.getElementById("tilepages-container");
    
    if (downTile.type == "hsplit") container.style.setProperty("cursor","url(" + chrome.runtime.getURL("img-cursor-ns.png") + ") 4 11,ns-resize","");
    else container.style.setProperty("cursor","url(" + chrome.runtime.getURL("img-cursor-ew.png") + ") 11 4,ew-resize","");
    
    container.addEventListener("mousemove",onMouseMoveSplitter,false);
    container.addEventListener("mouseup",onMouseUpSplitter,false);
    container.addEventListener("mouseleave",onMouseUpSplitter,false);
    container.addEventListener("dragstart",onDragStartSplitter,false);
}

function onMouseMoveSplitter(event)
{
    var tile0,tile1,diffX,diffY,height,width,iframe;
    
    if (moveTimeout != null) return;  /* allow time for previous resizeTile/drawTile to complete */
    
    moveTimeout = window.setTimeout(function() { moveTimeout = null; },10);
    
    tile0 = downTile.childTiles[downSplitter];
    tile1 = downTile.childTiles[downSplitter+1];
    
    if (downTile.type == "hsplit")  /* horizontal split */
    {
        diffY = lastScreenY-event.screenY;
        height = tile0.height+tile1.height;
        
        if (tile0.height-diffY < splitterSnap && diffY > 0)
        {
            tile0.ratio = 0;
            tile1.ratio = height/(downTile.height-downTile.splitterIds.length*splitterSize);
            resizeTile(tile0,tile0.width,0,tile0.left,tile0.top);
            resizeTile(tile1,tile1.width,height,tile1.left,tile0.top+splitterSize);
        }
        else if (tile1.height+diffY < splitterSnap && diffY < 0)
        {
            tile0.ratio = height/(downTile.height-downTile.splitterIds.length*splitterSize);
            tile1.ratio = 0;
            resizeTile(tile1,tile1.width,0,tile1.left,tile0.top+height+splitterSize);
            resizeTile(tile0,tile0.width,height,tile0.left,tile0.top);
        }
        else
        {
            tile0.ratio = (tile0.height-diffY)/(downTile.height-downTile.splitterIds.length*splitterSize);
            tile1.ratio = (tile1.height+diffY)/(downTile.height-downTile.splitterIds.length*splitterSize);
            resizeTile(tile0,tile0.width,tile0.height-diffY,tile0.left,tile0.top);
            resizeTile(tile1,tile1.width,tile1.height+diffY,tile1.left,tile1.top-diffY);
        }
    }
    else  /* vertical split */
    {
        diffX = lastScreenX-event.screenX;
        width = tile0.width+tile1.width; 
        
        if (tile0.width-diffX < splitterSnap && diffX > 0)
        {
            tile0.ratio = 0;
            tile1.ratio = width/(downTile.width-downTile.splitterIds.length*splitterSize);
            resizeTile(tile0,0,tile0.height,tile0.left,tile0.top);
            resizeTile(tile1,width,tile1.height,tile0.left+splitterSize,tile1.top);
        }
        else if (tile1.width+diffX < splitterSnap && diffX < 0)
        {
            tile0.ratio = width/(downTile.width-downTile.splitterIds.length*splitterSize);
            tile1.ratio = 0;
            resizeTile(tile1,0,tile1.height,tile0.left+width+splitterSize,tile1.top);
            resizeTile(tile0,width,tile0.height,tile0.left,tile0.top);
        }
        else
        {
            tile0.ratio = (tile0.width-diffX)/(downTile.width-downTile.splitterIds.length*splitterSize);
            tile1.ratio = (tile1.width+diffX)/(downTile.width-downTile.splitterIds.length*splitterSize);
            resizeTile(tile0,tile0.width-diffX,tile0.height,tile0.left,tile0.top);
            resizeTile(tile1,tile1.width+diffX,tile1.height,tile1.left-diffX,tile1.top);
        }
    }
    
    drawTile(downTile,true);
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
    
    lastScreenX = event.screenX;
    lastScreenY = event.screenY;
}

function onMouseUpSplitter(event)
{
    var i,iframes,iframe,tile,container;
    
    iframes = document.querySelectorAll("[name^=tilepages-panel-]");
    
    for (i = 0; i < iframes.length; i++)
    {
        iframe = document.getElementById(iframes[i].name);
        
        tile = findTileByPanel(rootTile,iframes[i].name);
        
        if (tile.scriptLoaded) iframe.style.removeProperty("pointer-events");
    }
    
    container = document.getElementById("tilepages-container");
    
    container.style.removeProperty("cursor");
    
    container.removeEventListener("mousemove",onMouseMoveSplitter,false);
    container.removeEventListener("mouseup",onMouseUpSplitter,false);
    container.removeEventListener("mouseleave",onMouseUpSplitter,false);
    container.removeEventListener("dragstart",onDragStartSplitter,false);
    
    saveRestoreLayout();
}

function onDragStartSplitter(event)
{
    event.preventDefault();
    event.stopPropagation();
}

function onDblClickSplitter(event)
{
    var tile;

    tile = findTileBySplitter(rootTile,event.currentTarget.id);
    
    equalizeTile(tile);
}

/************************************************************************/

/* Panel container event functions */

function onMouseDownContainer(event)
{
    var tile;
    
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    tile = findTileByClientCoords(rootTile,event.clientX,event.clientY);
    
    if (tile != null) onMouseDownIframe(tile.panelId);
    
    event.preventDefault();
}

function onDblClickContainer(event)
{
    var tile;
    
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    tile = findTileByClientCoords(rootTile,event.clientX,event.clientY);
    
    if (tile != null) onDblClickIframe(tile.panelId);
}

function onDragOverContainer(event)
{
    var tile;
    
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    tile = findTileByClientCoords(rootTile,event.clientX,event.clientY);
    
    if (tile != null) onDragOverIframe(tile.panelId,event.shiftKey,event.clientX-tile.left,event.clientY-tile.top,event.shiftKey);
    
    event.preventDefault();
}

function onDragLeaveContainer(event)
{
    var img;
    
    img = document.getElementById("tilepages-indicator");
    
    if (img != null) document.getElementById("tilepages-container").removeChild(img);
    
    event.preventDefault();
}

function onDropContainer(event)
{
    var tile;
    
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    if (event.dataTransfer.getData("text/plain") != "") dragText = event.dataTransfer.getData("text/plain");
    
    tile = findTileByClientCoords(rootTile,event.clientX,event.clientY);
    
    if (tile != null) onDropIframe(tile.panelId,qualifyUrl(dragText),event.shiftKey,event.clientX-tile.left,event.clientY-tile.top);
    
    event.preventDefault();
}

/************************************************************************/

/* Navigation bar event functions */

function onDragOverNavBar(event)
{
    var tile;
    
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    event.preventDefault();
}

function onDropNavBar(event)
{
    var urlbar;
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = event.dataTransfer.getData("text/plain");
    
    goToNewUrl(event);
    
    event.preventDefault();
}

function onClickBackward(event)
{
    if (activeTile.historyIndex > 0)
    {
        activeTile.historyIndex--;
        
        goToHistoryUrl();
    }
}

function onClickForward(event)
{
    if (activeTile.historyIndex < activeTile.historyUrl.length-1)
    {
        activeTile.historyIndex++;
        
        goToHistoryUrl();
    }
}

function onClickReload(event)
{
    goToHistoryUrl();
}

function onFocusUrlBar(event)
{
    var urlbar;
    
    if (urlbarFocused) return;
    
    urlbarFocused = true;
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.select();
}

function onBlurUrlBar(event)
{
    urlbarFocused = false;    
}

function onInputUrlBar(event)
{
    var urlbar,button;
    
    urlbar = document.getElementById("tilepages-urlbar");
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
}

function onDropUrlBar(event)
{
    var urlbar;
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = event.dataTransfer.getData("text/plain");
    
    goToNewUrl(event);
    
    event.preventDefault();
}

function onKeypressUrlBar(event)
{
    if (event.keyCode == 13)
    {
        onClickGo(event);
    }
}

function onClickGo(event)
{
    var i,urlbar,datalist,option;
    
    if (!layoutTabIncognito)
    {
        urlbar = document.getElementById("tilepages-urlbar");
        datalist = document.getElementById("tilepages-suggestions");
        
        for (i = 0; i < urlbarSuggestions.length; i++)
            if (urlbar.value == urlbarSuggestions[i]) break;
        
        if (i == urlbarSuggestions.length)
        {
            urlbarSuggestions.push(urlbar.value);
            
            option = document.createElement("option");
            option.setAttribute("value",urlbar.value);
            datalist.appendChild(option);
            
            if (urlbarSuggestions.length > 50)
            {
                urlbarSuggestions.shift();
                
                option = datalist.childElements[0];
                datalist.removeChild(option);
            }
            
            chrome.storage.local.set({ "urlbar-suggestions": urlbarSuggestions });
        }
    }
    
    goToNewUrl(event);
}

function onClickHome(event)
{
    var urlbar;
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = (homePage != "") ? homePage : "about:blank";
    
    goToNewUrl();
}

function onClickBookmark(event)
{
    var urlbar,titlebar;
    
    urlbar = document.getElementById("tilepages-urlbar");
    titlebar = document.getElementById("tilepages-titlebar");
    
    chrome.runtime.sendMessage({ type: "createBookmark", url: urlbar.value , title: titlebar.value });
}

function onClickExpandTile(event)
{
    if (viewTiled) expandTile(activeTile);
}

function onClickEqualizeTile(event)
{
    if (viewTiled) equalizeTile(activeTile.parentTile);
}

function onClickUntilePage(event)
{
    if (viewTiled) untilePage(activeTile);
}

function onClickOpenInTab(event)
{
    chrome.tabs.create({ url: activeTile.historyUrl[activeTile.historyIndex], active: true });
}

function onClickSyncScroll(event)
{
    toggleSync();
}

function onClickViewTiled(event)
{
    toggleView();
}

function goToHistoryUrl()
{
    var urlbar,titlebar,iframe,button;
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = activeTile.historyUrl[activeTile.historyIndex];
    
    titlebar = document.getElementById("tilepages-titlebar");
    titlebar.value = activeTile.historyTitle[activeTile.historyIndex];
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.src = activeTile.historyUrl[activeTile.historyIndex];
    
    activeTile.scriptLoaded = false;
    iframe.style.setProperty("pointer-events","none","");
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
    
    button = document.getElementById("tilepages-backward");
    button.disabled = (activeTile.historyIndex <= 0);
    
    button = document.getElementById("tilepages-forward");
    button.disabled = (activeTile.historyIndex >= activeTile.historyUrl.length-1);
}

function goToNewUrl()
{
    var urlbar,titlebar,iframe,button;
    
    urlbar = document.getElementById("tilepages-urlbar");
    titlebar = document.getElementById("tilepages-titlebar");
    
    titlebar.value = "";
    
    iframe = document.getElementById(activeTile.panelId);
    iframe.src = qualifyUrl(urlbar.value);
    
    activeTile.scriptLoaded = false;
    iframe.style.setProperty("pointer-events","none","");
    
    if (urlbar.value != activeTile.historyUrl[activeTile.historyIndex])
    {
        activeTile.historyIndex++;
        activeTile.historyUrl.length = activeTile.historyIndex;
        activeTile.historyUrl[activeTile.historyIndex] = urlbar.value;
        activeTile.historyTitle.length = activeTile.historyIndex;
        activeTile.historyTitle[activeTile.historyIndex] = "";
    }
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
    
    button = document.getElementById("tilepages-backward");
    button.disabled = (activeTile.historyIndex <= 0);
    
    button = document.getElementById("tilepages-forward");
    button.disabled = (activeTile.historyIndex >= activeTile.historyUrl.length-1);
}

/************************************************************************/

/* Iframe internal event functions */

function onLoadIframe(panelId,url,title,frameId)
{
    var iframe,tile,urlbar,titlebar,button;
    
    iframe = document.getElementById(panelId);
    
    tile = findTileByPanel(rootTile,panelId);
    
    tile.frameId = frameId;
    
    tile.scriptLoaded = true;
    iframe.style.removeProperty("pointer-events");
    
    if (tile.scrollInit)
    {
        chrome.tabs.sendMessage(layoutTabId,{ type: "initScrollPosn", scrollx: tile.scrollX, scrolly: tile.scrollY },{ frameId: tile.frameId });
        
        tile.scrollInit = false;
    }
    
    urlbar = document.getElementById("tilepages-urlbar");
    titlebar = document.getElementById("tilepages-titlebar");
    
    if (tile == activeTile)
    {
        urlbar.value = url;
        titlebar.value = title;
    }
    
    tile.historyUrl[tile.historyIndex] = url;
    tile.historyTitle.length = tile.historyIndex;
    tile.historyTitle[tile.historyIndex] = title;
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
    
    button = document.getElementById("tilepages-backward");
    button.disabled = (activeTile.historyIndex <= 0);
    
    button = document.getElementById("tilepages-forward");
    button.disabled = (activeTile.historyIndex >= activeTile.historyUrl.length-1);
}

function onMouseDownIframe(panelId)
{
    var iframe,urlbar,titlebar,button;
    
    if (viewTiled)
    {
        iframe = document.getElementById(activeTile.panelId);
        iframe.style.setProperty("border",tileBorder + "px solid transparent","");
        
        iframe = document.getElementById(panelId);
        iframe.style.setProperty("border",tileBorder + "px solid #5070A0","");
    }
    
    activeTile = findTileByPanel(rootTile,panelId);
    
    focusActiveTile();
    
    urlbar = document.getElementById("tilepages-urlbar");
    urlbar.value = activeTile.historyUrl[activeTile.historyIndex];
    
    titlebar = document.getElementById("tilepages-titlebar");
    titlebar.value = activeTile.historyTitle[activeTile.historyIndex];
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
    
    button = document.getElementById("tilepages-backward");
    button.disabled = (activeTile.historyIndex <= 0);
    
    button = document.getElementById("tilepages-forward");
    button.disabled = (activeTile.historyIndex >= activeTile.historyUrl.length-1);
    
    chrome.runtime.sendMessage({ type: "updateContextMenu" });
}

function onClickIframe(panelId,nextprev,shift,url)
{
    var i,urlbar,titlebar,tile,iframe,button;
    
    urlbar = document.getElementById("tilepages-urlbar");
    titlebar = document.getElementById("tilepages-titlebar");
    
    tile = findTileByPanel(rootTile,panelId);
    
    if (nextprev)  /* open link in previous or next tile */
    {
        for (i = 0; i < tile.parentTile.childTiles.length; i++) if (tile.parentTile.childTiles[i] == tile) break;
        
        if (shift) i = (i == 0) ? tile.parentTile.childTiles.length-1 : i-1;  /* open in previous tile */
        else i = (i == tile.parentTile.childTiles.length-1) ? 0 : i+1;  /* open in next tile */
        
        tile = tile.parentTile.childTiles[i];
        
        iframe = document.getElementById(tile.panelId);
        iframe.src = url;
    }
    else  /* open link in this tile */
    {
        iframe = document.getElementById(panelId);
        iframe.src = url;
    }
    
    tile.scriptLoaded = false;
    iframe.style.setProperty("pointer-events","none","");
    
    if (tile == activeTile)
    {
        urlbar.value = url;
        titlebar.value = "";
    }
    
    if (url != tile.historyUrl[tile.historyIndex])
    {
        tile.historyIndex++;
        tile.historyUrl.length = tile.historyIndex;
        tile.historyUrl[tile.historyIndex] = url;
        tile.historyTitle.length = tile.historyIndex;
        tile.historyTitle[tile.historyIndex] = "";
    }
    
    updatePageDetails();
    
    saveRestoreLayout();
    
    button = document.getElementById("tilepages-go");
    button.disabled = (urlbar.value.length == 0);
    
    button = document.getElementById("tilepages-backward");
    button.disabled = (activeTile.historyIndex <= 0);
    
    button = document.getElementById("tilepages-forward");
    button.disabled = (activeTile.historyIndex >= activeTile.historyUrl.length-1);
}

function onDblClickIframe(panelId)
{
    var tile;
    
    tile = findTileByPanel(rootTile,panelId);
    
    expandTile(tile);
}

function onScrollIframe(panelId,scrollX,scrollY)
{
    var i,tile,deltaX,deltaY;
    
    if (viewTiled)  /* tiled view */
    {
        tile = findTileByPanel(rootTile,panelId);
        
        deltaX = scrollX-tile.scrollX;
        deltaY = scrollY-tile.scrollY;
        
        tile.scrollX += deltaX;
        tile.scrollY += deltaY;
        
        if (syncScroll && (!syncActive || tile == activeTile))
        {
            for (i = 0; i < tile.parentTile.childTiles.length; i++)
            {
                if (tile.parentTile.childTiles[i] != tile && tile.parentTile.childTiles[i].type == "panel")
                {
                    tile.parentTile.childTiles[i].scrollX += deltaX;
                    tile.parentTile.childTiles[i].scrollY += deltaY;
                    
                    chrome.tabs.sendMessage(layoutTabId,{ type: "setScrollPosn",
                                                          scrollx: tile.parentTile.childTiles[i].scrollX,
                                                          scrolly: tile.parentTile.childTiles[i].scrollY },
                                                          { frameId: tile.parentTile.childTiles[i].frameId });
               }
            }
        }
        
        saveRestoreLayout();
    }
}

function onDragOverIframe(panelId,shift,clientX,clientY)
{
    var tile,direction,img;
    
    tile = findTileByPanel(rootTile,panelId);
    
    img = document.getElementById("tilepages-indicator");
    
    if (img != null) document.getElementById("tilepages-container").removeChild(img);
    
    if (shift)  /*  tile page */
    {
        direction = calcDirection(tile,clientX,clientY);
        
        img = document.createElement("img");
        img.src = chrome.runtime.getURL("img-" + direction + ".png");
        img.setAttribute("id","tilepages-indicator");
        img.style.setProperty("position","fixed","");
        img.style.setProperty("left",(tile.left+tile.width/2-25) + "px","");
        img.style.setProperty("top",(tile.top+tile.height/2-25) + "px","");
        img.style.setProperty("width","50px","");
        img.style.setProperty("height","50px","");
        img.style.setProperty("z-index","2","");
        img.style.setProperty("pointer-events","none","");
        document.getElementById("tilepages-container").appendChild(img);
    }
}

function onDropIframe(panelId,text,shift,clientX,clientY)
{
    var tile,direction,img;
    
    tile = findTileByPanel(rootTile,panelId);
    
    img = document.getElementById("tilepages-indicator");
    
    if (img != null) document.getElementById("tilepages-container").removeChild(img);
    
    if (shift)  /* tile page */
    {
        direction = calcDirection(tile,clientX,clientY);
        
        tilePage(tile,qualifyUrl(text),direction);
    }
    else  /* assign page */
    {
        assignPage(tile,qualifyUrl(text));
    }
}

function calcDirection(tile,clientX,clientY)
{
    var aspectRatio,remainderY,direction;
    
    aspectRatio = tile.width/tile.height;
    remainderY = tile.height-1-clientY;
    
    if (clientX >= clientY*aspectRatio && clientX < remainderY*aspectRatio) direction = "above";
    else if (clientX < clientY*aspectRatio && clientX < remainderY*aspectRatio) direction = "left";
    else if (clientX >= clientY*aspectRatio && clientX >= remainderY*aspectRatio) direction = "right";
    else if (clientX < clientY*aspectRatio && clientX >= remainderY*aspectRatio) direction = "below";
    
    return direction;
}

/************************************************************************/

/* Utility functions */

function findTileByPanel(tile,panelId)
{
    var i,value;
    
    if (tile.type == "panel")  /* panel tile */
    {
        if (tile.panelId == panelId) return tile;
    }
    else  /* split tile */
    {
        for (i = 0; i < tile.childTiles.length; i++)
            if ((value = findTileByPanel(tile.childTiles[i],panelId)) != null) return value;
    }
    
    return null;
}

function findTileBySplitter(tile,splitterId)
{
    var i,value;
    
    if (tile.type == "panel")  /* panel tile */
    {
    }
    else  /* split tile */
    {
        for (i = 0; i < tile.splitterIds.length; i++)
            if (tile.splitterIds[i] == splitterId) return tile;
        
        for (i = 0; i < tile.childTiles.length; i++)
            if ((value = findTileBySplitter(tile.childTiles[i],splitterId)) != null) return value;
    }
    
    return null;
}

function findTileByClientCoords(tile,clientX,clientY)
{
    var i,value;
    
    if (tile.type == "panel")  /* panel tile */
    {
        if (clientX >= tile.left && clientX <= tile.left+tile.width  && 
            clientY >= tile.top && clientY <= tile.top+tile.height) return tile;
        else return null;
    }
    else  /* split tile */
    {
        for (i = 0; i < tile.childTiles.length; i++)
            if ((value = findTileByClientCoords(tile.childTiles[i],clientX,clientY)) != null) return value;
    }
}

function focusActiveTile()
{
    var iframe;
    
    iframe = document.getElementById(activeTile.panelId);
    
    iframe.focus();
    
    lastActiveElement = iframe;
}

function saveRestoreLayout()
{
    saveLayout("",2);
}

function updatePageDetails()
{
    var i,iframes,tile;
    var pageUrls = new Array();
    var pageTitles = new Array();
    
    iframes = document.querySelectorAll("[name^=tilepages-panel-]");
    
    for (i = 0; i < iframes.length; i++)
    {
        tile = findTileByPanel(rootTile,iframes[i].name);
        
        pageUrls[i] = tile.historyUrl[tile.historyIndex];
        pageTitles[i] = tile.historyTitle[tile.historyIndex];
    }
    
    chrome.runtime.sendMessage({ type: "updatePageDetails", pageurls: pageUrls, pagetitles: pageTitles });
}

function allocateUrl(pageUrls,index)
{
    if (index < pageUrls.length) return pageUrls[index];
    else if (homePage != "") return homePage;
    else return "about:blank";
}

function restrictedPage(url)
{
    /* Pages that cannot be loaded into an <iframe> (1) */
    /* Pages that do not allow content scripts to be loaded (2) */
    /* Pages that are not allowed by WebExtensions Content Security Policy (3) */
    
    return (url.substr(0,6) == "about:" || url.substr(0,7) == "chrome:" ||  /*(1)*/
            url.substr(0,5) == "data:" || url.substr(0,11) == "javascript:" ||  /*(2)*/
            url.substr(0,14) == "moz-extension:" || url.substr(0,26) == "https://addons.mozilla.org" ||  /*(3)*/
            url.substr(0,17) == "chrome-extension:" || url.substr(0,34) == "https://chrome.google.com/webstore");  /*(3)*/
}

function qualifyUrl(url)
{
    var fields = new Array();
    var parts = new Array();
    
    /* URL syntax: scheme://domain[:port][/path][?query][#fragment] */
    
    /* fields[0] - full url  */
    /* fields[1] - scheme:// */
    /* fields[2] - domain    */  /* or hostname for file URL */
    /* fields[3] - :port     */
    /* fields[4] - /path     */
    /* fields[5] - ?query    */
    /* fields[6] - #fragment */
    
    /* domain - [subdomains]maindomain.publicsuffix */
    /* subdomains - subdomain.[subdomain.]...       */
    /* publicsuffix - suffix[.suffix]...            */
    
    if (url == "about:blank") return url;
    
    fields = url.match(/(.+:\/\/)?([^:/?#]*)?(:[^/?#]*)?(\/[^?#]*)?(\?[^#]*)?(#.*)?/);  /* allow scheme and domain not to match */
    
    if (fields == null) return false;  /* url not matched */
    
    fields[0] = "";
    
    if (fields[1] != "http://" && fields[1] != "https://") fields[1] = "http://";
    
    parts = fields[2].split(".");
    
    if (parts.length == 1) fields[2] = fields[2] + ".com";
    
    return fields.join("");
}

/************************************************************************/
