/************************************************************************/
/*                                                                      */
/*      Tile Pages WE - Generic WebExtension - Content Pages            */
/*                                                                      */
/*      Javascript for Content Pages                                    */
/*                                                                      */
/*      Last Edit - 10 Jan 2019                                         */
/*                                                                      */
/*      Copyright (C) 2016-2019 DW-dev                                  */
/*                                                                      */
/*      Distributed under the GNU General Public License version 2      */
/*      See LICENCE.txt file and http://www.gnu.org/licenses/           */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/* Refer to Google Chrome developer documentation:                      */
/*                                                                      */
/*  https://developer.chrome.com/extensions/content_scripts             */
/*  https://developer.chrome.com/extensions/messaging                   */
/*                                                                      */
/*  https://developer.chrome.com/extensions/match_patterns              */
/*                                                                      */
/*  https://developer.chrome.com/extensions/runtime                     */
/*  https://developer.chrome.com/extensions/storage                     */
/*                                                                      */
/************************************************************************/

/* Loaded into all frames (including iframes) of all content pages */

/************************************************************************/

/* Global variables */

var openLinks = 0;

var dragLimiter = 0;

var ignoreScrollEvent = false;

/************************************************************************/

/* Initialize on script load */

if (window.name.substr(0,16) == "tilepages-panel-")
{
    /* If <iframe> in Tile Pages WE container page */
    
    if (document.readyState != "loading") onLoadPage();
    else
    {
        window.addEventListener("DOMContentLoaded",
        function(event)
        {
            if (document.readyState != "loading") onLoadPage();
        },false);
    }
}

/************************************************************************/

/* Initialize on page load */

function onLoadPage()
{
    chrome.storage.local.get(null,
    function(object)
    {
        openLinks = object["options-openlinks"];
        
        /* Add listeners */
        
        addListeners();
        
        /* Tiled page loaded */
        
        chrome.runtime.sendMessage({ type: "iframeLoad", panelid: window.name, url: document.URL, title: document.title });
    });
}

/************************************************************************/

/* Add listeners */

function addListeners()
{
    /* Storage changed listener */
    
    chrome.storage.onChanged.addListener(
    function(changes,areaName)
    {
        chrome.storage.local.get(null,
        function(object)
        {
            openLinks = object["options-openlinks"];
        });
    });
    
    /* Add event listeners */
    
    document.addEventListener("mousedown",onMouseDown,true);
    document.addEventListener("click",onClick,true);
    document.addEventListener("dblclick",onDblClick,true);
    document.addEventListener("scroll",onScroll,true);
    document.addEventListener("dragstart",onDragStart,true);
    document.addEventListener("dragover",onDragOver,true);
    document.addEventListener("drop",onDrop,true);
    
    /* Message received listener */
    
    chrome.runtime.onMessage.addListener(
    function(message,sender,sendResponse)
    {
        switch (message.type)
        {
            /* Messages from script in container page */
            
            case "initScrollPosn":
                
                ignoreScrollEvent = true;
                
                window.scrollTo(message.scrollx,message.scrolly);

                break;
                
            case "setScrollPosn":
                
                ignoreScrollEvent = true;

                window.scrollTo(message.scrollx,message.scrolly);
                
                break;
        }
    });
}
                
/************************************************************************/

/* Mouse event functions */

function onMouseDown(event)
{
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    chrome.runtime.sendMessage({ type: "iframeMouseDown", panelid: window.name });
}

function onClick(event)
{
    var element;
    
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    element = event.target;
    while (element && !element.href) element = element.parentNode;
    
    if (element)
    {
        if ((openLinks == 1 && event.button == 0 && event.ctrlKey) ||
            (openLinks == 2 && event.button == 1 && !event.ctrlKey) ||
            (openLinks == 3 && event.button == 1 && event.ctrlKey))
        {
            chrome.runtime.sendMessage({ type: "iframeClick",  panelid: window.name, url: element.href, nextprev: true, shift: event.shiftKey });
        }
        else if (event.button != 2)
        {
            chrome.runtime.sendMessage({ type: "iframeClick", panelid: window.name, url: element.href, nextprev: false, shift: false });
        }
    }
            
    event.preventDefault();
    event.stopPropagation();
}

function onDblClick(event)
{
    var element,selection,range,rects;
    
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    element = event.target;
    while (element && !element.href) element = element.parentNode;
    
    selection = document.getSelection();
    range = selection.getRangeAt(0);
    rects = range.getClientRects();
    
    if (!element)  /* current target not inside link */
    {
        if (rects.length == 0 ||
            event.clientX < rects[0].left-10 || event.clientX > rects[0].right+10 ||
            event.clientY < rects[0].top-10 || event.clientY > rects[0].bottom+10)  /* cursor not inside selection */
        {
            selection.removeAllRanges();
            
            chrome.runtime.sendMessage({ type: "iframeDblClick", panelid: window.name });
        }
    }
}

function onScroll(event)
{
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    if (!ignoreScrollEvent) chrome.runtime.sendMessage({ type: "iframeScroll", panelid: window.name, scrollx: window.scrollX, scrolly: window.scrollY });
    
    ignoreScrollEvent = false;
}

function onDragStart(event)
{
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    chrome.runtime.sendMessage({ type: "iframeDragStart", panelid: window.name, text: event.dataTransfer.getData("text/plain") });
}

function onDragOver(event)
{
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    dragLimiter++;
    
    if (dragLimiter > 10)
    {
        dragLimiter = 0;
        
        chrome.runtime.sendMessage({ type: "iframeDragOver",  panelid: window.name, shift: event.shiftKey, clientX: event.clientX, clientY: event.clientY });
    }
    
    event.preventDefault();
}

function onDrop(event)
{
    if (typeof event.isTrusted != "undefined" && !event.isTrusted) return;
    
    chrome.runtime.sendMessage({ type: "iframeDrop", panelid: window.name, text: event.dataTransfer.getData("text/plain"), shift: event.shiftKey, clientX: event.clientX, clientY: event.clientY });
    
    event.preventDefault();
}

/************************************************************************/
