/************************************************************************/
/*                                                                      */
/*      Tile Pages WE - Generic WebExtension - Options Page             */
/*                                                                      */
/*      Javascript for Options Page                                     */
/*                                                                      */
/*      Last Edit - 20 Jan 2019                                         */
/*                                                                      */
/*      Copyright (C) 2016-2019 DW-dev                                  */
/*                                                                      */
/*      Distributed under the GNU General Public License version 2      */
/*      See LICENCE.txt file and http://www.gnu.org/licenses/           */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Refer to Google Chrome developer documentation:                     */
/*                                                                      */
/*  https://developer.chrome.com/extensions/optionsV2                   */
/*                                                                      */
/*  https://developer.chrome.com/extensions/storage                     */
/*                                                                      */
/************************************************************************/

"use strict";

/************************************************************************/

/* Global variables */

var isFirefox;
var ffVersion;

var platformOS;

/************************************************************************/

/* Listener for options page load */

document.addEventListener("DOMContentLoaded",onLoadPage,false);

/************************************************************************/

/* Load options */

function onLoadPage(event)
{
    /* Load options from local storage */
    
    chrome.storage.local.get(null,
    function(object)
    {
        /* Load environment */
        
        isFirefox = object["environment-isfirefox"];
        
        if (isFirefox) ffVersion = object["environment-ffversion"];
        
        platformOS = object["environment-platformos"];
        
        if (isFirefox && ffVersion >= 60) document.body.setAttribute("shortcuts","");
        
        /* General options */
        
        document.getElementById("options-shownavbar").checked = object["options-shownavbar"];
        document.getElementById("options-settitle").checked = object["options-settitle"];
        document.getElementById("options-showsubmenu").checked = object["options-showsubmenu"];
        
        document.getElementById("options-assigntabs").checked = object["options-assigntabs"];
        document.getElementById("options-assignleft").checked = object["options-assignleft"];
        
        document.getElementById("options-startaction").elements["action"].value = object["options-startaction"];
		
        document.getElementById("options-assignleft").disabled = !document.getElementById("options-assigntabs").checked;
        
		/* Page Settings options */
		
        document.getElementById("options-homepage").value = object["options-homepage"];
        
        document.getElementById("options-openlinks").elements["clicks"].value = object["options-openlinks"];
        
        document.getElementById("options-syncactive").checked = object["options-syncactive"];
        
		/* Advanced options */
        
        document.getElementById("options-xframeoptions").checked = object["options-xframeoptions"];
        document.getElementById("options-cspframeancestors").checked = object["options-cspframeancestors"];
		
        /* Keyboard shortcuts */
        
        if (isFirefox && ffVersion >= 60)
        {
            chrome.commands.getAll(
            function(commands)
            {
                var i;
                
                for (i = 0; i < commands.length; i++)
                {
                    console.log(commands[i].name + " - " + commands.length);
                    if (commands[i].name == "newcloselayout") document.getElementById("options-shortcuts-newcloselayout").value = commands[i].shortcut;
                    else if (commands[i].name == "untilepage") document.getElementById("options-shortcuts-untilepage").value = commands[i].shortcut;
                    else if (commands[i].name == "syncscroll") document.getElementById("options-shortcuts-syncscroll").value = commands[i].shortcut;
                    else if (commands[i].name == "viewtiled") document.getElementById("options-shortcuts-viewtiled").value = commands[i].shortcut;
                }
            });
        }
    });
    
    /* Add listeners for click on tab buttons */
    
    document.getElementById("options-tabbar-general").addEventListener("click",showGeneralTab,false);
    document.getElementById("options-tabbar-pagesettings").addEventListener("click",showPageSettingsTab,false);
    document.getElementById("options-tabbar-advanced").addEventListener("click",showAdvancedTab,false);
    document.getElementById("options-tabbar-shortcuts").addEventListener("click",showShortcutsTab,false);
    
    /* Add listeners for click on assign tabs checkbox */
    
    document.getElementById("options-assigntabs").addEventListener("click",onClickAssignTabs,false);
    
    /* Add listener for click on save button */
    
    document.getElementById("options-save-button").addEventListener("click",onClickSave,false);
    
    /* Add listener for click on reset all button */
    
    document.getElementById("options-resetall-button").addEventListener("click",onClickResetAll,false);
    
    /* Wait for page layout to complete */
    
    document.getElementById("options").style.setProperty("opacity","0","");
    
    window.setTimeout(
    function()
    {
        var width1,width2,width3,width4,height1,height2,height3,height4;
        
        /* Equalize widths of tabs */
        
        width1 = window.getComputedStyle(document.getElementById("options-tab-general"),null).getPropertyValue("width");
        width2 = window.getComputedStyle(document.getElementById("options-tab-pagesettings"),null).getPropertyValue("width");
        width3 = window.getComputedStyle(document.getElementById("options-tab-advanced"),null).getPropertyValue("width");
        width4 = window.getComputedStyle(document.getElementById("options-tab-shortcuts"),null).getPropertyValue("width");
        
        width1 = width1.substr(0,width1.length-2);
        width2 = width2.substr(0,width2.length-2);
        width3 = width3.substr(0,width3.length-2);
        width4 = width4.substr(0,width4.length-2);
        
        width1 = Math.max(width1,width2,width3,width4);
        
        document.getElementById("options-tab-general").style.setProperty("width",width1 + "px","");
        document.getElementById("options-tab-pagesettings").style.setProperty("width",width1 + "px","");
        document.getElementById("options-tab-advanced").style.setProperty("width",width1 + "px","");
        document.getElementById("options-tab-shortcuts").style.setProperty("width",width1 + "px","");
        
        /* Equalize heights of tabs */
        
        height1 = window.getComputedStyle(document.getElementById("options-tab-general"),null).getPropertyValue("height");
        height2 = window.getComputedStyle(document.getElementById("options-tab-pagesettings"),null).getPropertyValue("height");
        height3 = window.getComputedStyle(document.getElementById("options-tab-advanced"),null).getPropertyValue("height");
        height4 = window.getComputedStyle(document.getElementById("options-tab-shortcuts"),null).getPropertyValue("height");
        
        height1 = height1.substr(0,height1.length-2);
        height2 = height2.substr(0,height2.length-2);
        height3 = height3.substr(0,height3.length-2);
        height4 = height4.substr(0,height4.length-2);
        
        height1 = Math.max(height1,height2,height3,height4);
        
        document.getElementById("options-tab-general").style.setProperty("height",height1 + "px","");
        document.getElementById("options-tab-pagesettings").style.setProperty("height",height1 + "px","");
        document.getElementById("options-tab-advanced").style.setProperty("height",height1 + "px","");
        document.getElementById("options-tab-shortcuts").style.setProperty("height",height1 + "px","");
        
        /* Show general tab */
        
        showGeneralTab();
        
        document.getElementById("options").style.setProperty("opacity","1","");
    },50);
}

/************************************************************************/

/* Select tab */

function showGeneralTab(event)
{
    document.getElementById("options-tabbar-general").setAttribute("selected","");
    document.getElementById("options-tabbar-pagesettings").removeAttribute("selected");
    document.getElementById("options-tabbar-advanced").removeAttribute("selected");
    document.getElementById("options-tabbar-shortcuts").removeAttribute("selected");
    
    document.getElementById("options-tab-general").style.setProperty("display","block","");
    document.getElementById("options-tab-pagesettings").style.setProperty("display","none","");
    document.getElementById("options-tab-advanced").style.setProperty("display","none","");
    document.getElementById("options-tab-shortcuts").style.setProperty("display","none","");
}

function showPageSettingsTab(event)
{
    document.getElementById("options-tabbar-general").removeAttribute("selected");
    document.getElementById("options-tabbar-pagesettings").setAttribute("selected","");
    document.getElementById("options-tabbar-advanced").removeAttribute("selected");
    document.getElementById("options-tabbar-shortcuts").removeAttribute("selected");
    
    document.getElementById("options-tab-general").style.setProperty("display","none","");
    document.getElementById("options-tab-pagesettings").style.setProperty("display","block","");
    document.getElementById("options-tab-advanced").style.setProperty("display","none","");
    document.getElementById("options-tab-shortcuts").style.setProperty("display","none","");
}

function showAdvancedTab(event)
{
    document.getElementById("options-tabbar-general").removeAttribute("selected");
    document.getElementById("options-tabbar-pagesettings").removeAttribute("selected");
    document.getElementById("options-tabbar-advanced").setAttribute("selected","");
    document.getElementById("options-tabbar-shortcuts").removeAttribute("selected");
    
    document.getElementById("options-tab-general").style.setProperty("display","none","");
    document.getElementById("options-tab-pagesettings").style.setProperty("display","none","");
    document.getElementById("options-tab-advanced").style.setProperty("display","block","");
    document.getElementById("options-tab-shortcuts").style.setProperty("display","none","");
}

function showShortcutsTab(event)
{
    document.getElementById("options-tabbar-general").removeAttribute("selected");
    document.getElementById("options-tabbar-pagesettings").removeAttribute("selected");
    document.getElementById("options-tabbar-advanced").removeAttribute("selected");
    document.getElementById("options-tabbar-shortcuts").setAttribute("selected","");
    
    document.getElementById("options-tab-general").style.setProperty("display","none","");
    document.getElementById("options-tab-pagesettings").style.setProperty("display","none","");
    document.getElementById("options-tab-advanced").style.setProperty("display","none","");
    document.getElementById("options-tab-shortcuts").style.setProperty("display","block","");
}

/************************************************************************/

/* Enable or Disable options */

function onClickAssignTabs(event)
{
    document.getElementById("options-assignleft").disabled = !document.getElementById("options-assigntabs").checked;
}

/************************************************************************/

/* Save options */

function onClickSave(event)
{
    chrome.storage.local.set(
    {
        /* General options */
        
        "options-shownavbar": document.getElementById("options-shownavbar").checked,
        "options-settitle": document.getElementById("options-settitle").checked,
        "options-showsubmenu": document.getElementById("options-showsubmenu").checked,
        
        "options-assigntabs": document.getElementById("options-assigntabs").checked,
        "options-assignleft": document.getElementById("options-assignleft").checked,
        
        "options-startaction": document.getElementById("options-startaction").elements["action"].value,
		
		/* Page Settings options */
		
        "options-homepage": document.getElementById("options-homepage").value,
        
        "options-openlinks": +document.getElementById("options-openlinks").elements["clicks"].value,
        
        "options-syncactive": document.getElementById("options-syncactive").checked,
		
		/* Advanced options */
        
        "options-xframeoptions": document.getElementById("options-xframeoptions").checked,
        "options-cspframeancestors": document.getElementById("options-cspframeancestors").checked
    });
    
    /* Keyboard shortcuts */
    
    if (isFirefox && ffVersion >= 60)
    {
        try
        {
            chrome.commands.update({ name: "newcloselayout", shortcut: document.getElementById("options-shortcuts-newcloselayout").value });
        }
        catch (e)
        {
            chrome.commands.reset("newcloselayout");
            document.getElementById("options-shortcuts-newcloselayout").value = (platformOS == "mac") ? "MacCtrl+Shift+Up" : "Alt+Shift+Up";
        }
        
        try
        {
            chrome.commands.update({ name: "untilepage", shortcut: document.getElementById("options-shortcuts-untilepage").value });
        }
        catch (e)
        {
            chrome.commands.reset("untilepage");
            document.getElementById("options-shortcuts-untilepage").value = (platformOS == "mac") ? "MacCtrl+Shift+Down" : "Alt+Shift+Down";
        }
        
        try
        {
            chrome.commands.update({ name: "syncscroll", shortcut: document.getElementById("options-shortcuts-syncscroll").value });
        }
        catch (e)
        {
            chrome.commands.reset("syncscroll");
            document.getElementById("options-shortcuts-syncscroll").value = (platformOS == "mac") ? "MacCtrl+Shift+Left" : "Alt+Shift+Left";
        }
        
        try
        {
            chrome.commands.update({ name: "viewtiled", shortcut: document.getElementById("options-shortcuts-viewtiled").value });
        }
        catch (e)
        {
            chrome.commands.reset("viewtiled");
            document.getElementById("options-shortcuts-viewtiled").value = (platformOS == "mac") ? "MacCtrl+Shift+Right" : "Alt+Shift+Right";
        }
    }
    
    /* Display saved status for short period */
    
    document.getElementById("options-save-button").innerText = "Saved";
    document.getElementById("options-save-button").style.setProperty("font-weight","bold","");
    
    window.setTimeout(function()
    {
        document.getElementById("options-save-button").innerText = "Save";
        document.getElementById("options-save-button").style.setProperty("font-weight","normal","");
    }
    ,1000);
}

/************************************************************************/

/* Reset All options */

function onClickResetAll(event)
{
    /* General options */
    
    document.getElementById("options-shownavbar").checked = true;
    document.getElementById("options-settitle").checked = false;
    document.getElementById("options-showsubmenu").checked = true;
    
    document.getElementById("options-assigntabs").checked = true;
    document.getElementById("options-assignleft").checked = false;
    
    document.getElementById("options-startaction").elements["action"].value = 0;
	
    document.getElementById("options-assignleft").disabled = false;
    
	/* Page Settings options */
	
    document.getElementById("options-homepage").value = "";
	
    document.getElementById("options-openlinks").elements["clicks"].value = 3;
	
    document.getElementById("options-syncactive").checked = false;
	
	/* Advanced options */
	
	document.getElementById("options-xframeoptions").checked = false;
	document.getElementById("options-cspframeancestors").checked = false;
	
    /* Keyboard shortcuts */
    
    if (isFirefox && ffVersion >= 60)
    {
        document.getElementById("options-shortcuts-newcloselayout").value = (platformOS == "mac") ? "MacCtrl+Shift+Up" : "Alt+Shift+Up";
        document.getElementById("options-shortcuts-untilepage").value = (platformOS == "mac") ? "MacCtrl+Shift+Down" : "Alt+Shift+Down";
        document.getElementById("options-shortcuts-syncscroll").value = (platformOS == "mac") ? "MacCtrl+Shift+Left" : "Alt+Shift+Left";
        document.getElementById("options-shortcuts-viewtiled").value = (platformOS == "mac") ? "MacCtrl+Shift+Right" : "Alt+Shift+Right";
    }
    
    /* Display reset status for short period */
    
    document.getElementById("options-resetall-button").innerText = "Reset";
    document.getElementById("options-resetall-button").style.setProperty("font-weight","bold","");
    
    window.setTimeout(function()
    {
        document.getElementById("options-resetall-button").innerText = "Reset All";
        document.getElementById("options-resetall-button").style.setProperty("font-weight","normal","");
    }
    ,1000);
}

/************************************************************************/
